import os, random, pygame
from pygame.locals import *

if not pygame.font: print 'Warning, fonts disabled'
if not pygame.mixer: print 'Warning, sound disabled'

#
#Tile.py
#############################

class Tile(pygame.sprite.Sprite):
	tileType = "rock"
	tileImage = None
	tileRect = None

	def __init__(self, t, img):
		pygame.sprite.Sprite.__init__(self)
		self.tileType = t
		self.tileImage = img
		self.tileRect = img.get_rect()

	def clone(self):
		return Tile(self.tileType, self.tileImage)
	

#
#World.py
#############################
class World:
	orientation = "north"
	depth = 0
	width = 0
	height = 0
	windowSize = None
	
	tileWidth = 101
	tileDepth = 80
	tileHeight = 40
	
	shadowImages = None
	shadows = None
	
	terrain = [] # One-dimensional array, length == depth * width.
				 # Width varies faster.
				 # Each position has an array representing height.
	
	selectorPosition = (0, 0)
	
	def __init__(self, depth, width, height, shadowImages, terrain = []):
		self.depth = depth
		self.width = width
		self.height = height
		self.shadowImages = shadowImages
		
		if terrain == []:
			for i in range(depth*width):
				self.terrain.append([])
		else:
			for i in range(depth*width):
				self.terrain.append(terrain[i])
		
		self.windowSize = (self.tileWidth*width, self.tileHeight*height + self.tileDepth*(depth+1))
		self.updateShadows()

	def getPosition(self, x, y):
		return x*self.width + y
	
	def updateShadows(self):
		"""For the current terrain configuration detect the shadows inciding in
		   each position. Only to be called when modifying terrain or rotating it"""
		# FIXME: Will have to change this to adjust to different orientations
		terrainIndices = [(x, y) for y in range(self.depth) for x in range(self.width)]
		
		nw = map(self.getTileHeight, map(lambda x: self.getCorner(x, (-1, -1)), terrainIndices))
		n  = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 0, -1)), terrainIndices))
		ne = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 1, -1)), terrainIndices))
		w  = map(self.getTileHeight, map(lambda x: self.getCorner(x, (-1,  0)), terrainIndices))
		e  = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 1,  0)), terrainIndices))
		sw = map(self.getTileHeight, map(lambda x: self.getCorner(x, (-1,  1)), terrainIndices))
		s  = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 0,  1)), terrainIndices))
		se = map(self.getTileHeight, map(lambda x: self.getCorner(x, ( 1,  1)), terrainIndices))
		self.shadows = map(self.detectShadows, map(self.getTileHeight, terrainIndices), nw, n, ne, w, e, sw, s, se, terrainIndices)

	def getTileHeight(self, p):
		"""Returns the height of a tile space. Returns 0 if p is None"""
		if p == None:
			return 0
		return len(self.terrain[p[1]*self.width+p[0]])
	
	def getCorner(self, pos, offset):
		"""For a ordered pair pos return its position shifted by offset. If
		   the position falls outside of the terrain returns None"""
		if pos[0] + offset[0] < 0 or pos[1] + offset[1] < 0 or \
		   pos[0] + offset[0] >= self.width or pos[1] + offset[1] >= self.depth:
			return None
		else:
			return (pos[0] + offset[0], pos[1] + offset[1])
	
	def detectShadows(self, p, nw, n, ne, w, e, sw, s, se, ind):
		"""returns for each p the list of shadows it has to apply to p"""
		shadowSquare = dict()
		
		shadowSquare['nw'] = nw > p and w <= p and n <= p
		shadowSquare['n'] = n > p
		shadowSquare['ne'] = ne > p and e <= p and n <= p
		shadowSquare['w'] = w > p
		shadowSquare['e'] = e > p
		shadowSquare['sw'] = sw > p and w <= p
		shadowSquare['s'] = s > p
		shadowSquare['se'] = se > p and e <= p
		shadowSquare['sidew'] = sw >= p
		shadowSquare['topmost'] = p == self.height
		
		return [self.shadowImages[key] for key in shadowSquare.keys() if shadowSquare[key]]

	def draw(self, screen, offset = (0, 0)):
		#This is for drawing for "north" orientation only
		for i in range(self.depth):
			for j in range(self.width):
				tileSpace = self.terrain[i*(self.width)+j]
				shadow = self.shadows[i*(self.width)+j]
				for k in range(len(tileSpace)):
					tile = tileSpace[k]
					x = offset[0] + j*self.tileWidth
					y = offset[1] + i*self.tileDepth+(self.height - k)*self.tileHeight
					screen.blit(tile.tileImage, (x, y))
					if k+1 == len(tileSpace):
						for l in shadow:
							screen.blit(l, (x, y))

def generateRandomTerrain(d, w, h, tiles):
	terr = []

	for i in range(d*w):
		tileSpace = []
		for j in range(0, random.randint(1, h)):
			tileSpace.append(random.choice(tiles).clone())
		terr.append(tileSpace)
	return terr

def generatePlainTerrain(d, w, h, tile):
	terr = []

	for i in range(d*w):
		tileSpace = []
		for j in range(h):
			tileSpace.append(tile.clone())
		terr.append(tileSpace)
	return terr

#
#WorldScrollView.py
#############################
class WorldScrollView:
	view = None
	__viewPosition = (0, 0)
	__windowSize = None
	world = None
	border = (20, 30)
	
	def __init__(self, windowSize, world):
		self.view = pygame.Surface(windowSize, pygame.SRCALPHA, 32);
		self.__windowSize = windowSize
		self.__viewPosition = (0, 0)
		self.world = world
		self.update()
	
	def update(self):
		self.view.fill((0, 0, 0, 0))
		self.world.draw(self.view, self.__viewPosition)
	
	def draw(self, screen, viewPosition = (0, 0)):
		screen.blit(self.view, viewPosition)
		
	
	def getPosition(self):
		return self.__viewPosition*-1
	
	def setPosition(self, pos):
		self.__viewPosition = (-max(0, min(self.world.windowSize[0]+self.border[0]*2-self.__windowSize[0], pos[0])),
		                       -max(0, min(self.world.windowSize[1]+self.border[1]*2-self.__windowSize[1], pos[1])))

#
# CuteGod.py
###################################

def load_image(name, colorkey=None):
	fullname = name
	try:
		image = pygame.image.load(fullname)
	except pygame.error, message:
		print 'Cannot load image:', fullname
		raise SystemExit, message
	image = image.convert_alpha()
	if colorkey is not None:
		if colorkey is -1:
			colorkey = image.get_at((0,0))
		image.set_colorkey(colorkey, RLEACCEL)
	return image

def initShadowImages():
	shadowImages = dict()
	shadowImages['nw'] = load_image(os.path.join('data', 'Shadow North West.png'))
	shadowImages['n'] = load_image(os.path.join('data', 'Shadow North.png'))
	shadowImages['ne'] = load_image(os.path.join('data', 'Shadow North East.png'))
	shadowImages['w'] = load_image(os.path.join('data', 'Shadow West.png'))
	shadowImages['e'] = load_image(os.path.join('data', 'Shadow East.png'))
	shadowImages['sw'] = load_image(os.path.join('data', 'Shadow South West.png'))
	shadowImages['s'] = load_image(os.path.join('data', 'Shadow South.png'))
	shadowImages['se'] = load_image(os.path.join('data', 'Shadow South East.png'))
	shadowImages['sidew'] = load_image(os.path.join('data', 'Shadow Side West.png'))
	shadowImages['topmost'] = shadowImages['sidew']
	return shadowImages

def main():
	pygame.init()
	screen = pygame.display.set_mode( (800, 600) )
	clock = pygame.time.Clock()
	shadowImages = initShadowImages()

	backgroundImage = load_image(os.path.join('data', 'background.png'))
	plainTile = Tile('rock', load_image(os.path.join('data', 'Plain Block.png') ))
	grassTile = Tile('grass', load_image(os.path.join('data' , 'Grass Block.png') ))
	stoneTile = Tile('stone', load_image(os.path.join('data' , 'Stone Block.png') ))
	waterTile = Tile('water', load_image(os.path.join('data' , 'Water Block.png') ))

	#world = World(6, 8, 5, generatePlainTerrain(6, 8, 1, plainTile))
	world = World(4, 6, 5, shadowImages, generateRandomTerrain(4, 6, 3, [plainTile, grassTile, stoneTile, waterTile]))
	#world = World(3, 3, 5, shadowImages, [[stoneTile],[stoneTile,stoneTile],[stoneTile],[stoneTile],[stoneTile,stoneTile],[stoneTile],[stoneTile,stoneTile],[stoneTile],[stoneTile]])
	worldScrollView = WorldScrollView((600, 600), world)
	
	while 1:          
		screen.blit(backgroundImage, (0, 0))
		worldScrollView.draw(screen, (200, 0))
		worldScrollView.setPosition((0, 50))
		worldScrollView.update()
		pygame.display.flip()

		for event in pygame.event.get():
			if event.type == QUIT:
				return
			elif event.type == KEYDOWN and event.key == K_ESCAPE:
				return

		clock.tick(60)
	
if __name__ == '__main__':
	main()

